import json
import html
import os
from IPython.display import display_html


class Item:
    def __init__(self, data, id):
        self._data = data
        data['id'] = id
        data['geometry'] = []
        data['results'] = {}

    def add_result(self, key, value):
        results = self._data['results']

        results[key] = value

    def vec(self, v):
        return {'x': float(v[0]), 'y': float(v[1]), 'z': float(v[2])}

    def set_label_location(self, ref, act):
        self._data['ref_label_location'] = self.vec(ref)
        self._data['act_label_location'] = self.vec(act)

    def add_point(self, location, color, layer):
        self._data['geometry'].append({
            'type': 'point',
            'location': self.vec(location),
            'layer': layer,
            'color': color,
        })

    def add_line(self, points, color, layer):
        self._data['geometry'].append({
            'type': 'line',
            'points': [self.vec(point) for point in points],
            'layer': layer,
            'color': color,
        })

    def add_arrow(self, location, direction, color, layer):
        self._data['geometry'].append({
            'type': 'arrow',
            'location': self.vec(location),
            'direction': self.vec(direction),
            'layer': layer,
            'color': color,
        })

    def add_support(self, location, direction, color, layer):
        self._data['geometry'].append({
            'type': 'support',
            'location': self.vec(location),
            'direction': direction,
            'layer': layer,
            'color': color,
        })

    def spring(self, *args, **kwargs):
        pass


class Canvas3D:
    def __init__(self, height=600):
        self.settings = dict()
        self.settings['timestep'] = 0

        self.data = dict()
        self.data['timesteps'] = []

    def _embed_js(self, template, filename):
        path = os.path.join(os.path.dirname(__file__), 'html', filename)
        with open(path, 'r', encoding='UTF-8') as file:
            content = file.read()
            template = template.replace(f'<script type="text/javascript" src="{filename}"></script>', f'<script type="text/javascript">{content}</script>')
        return template

    def html(self, height, model):
        content = self.raw_html(height, model)

        return f'<iframe seamless frameborder="0" allowfullscreen width="100%" height="{height}" srcdoc="{html.escape(content)}"></iframe>'

    def raw_html(self, height, model):
        timesteps_data = []

        for model in model.get_model_history():
            nodes_data = []

            for node in model.nodes:
                node_data = {}
                item = Item(node_data, node.id)
                node.draw(item)
                nodes_data.append(node_data)

            elements_data = []

            for element in model.elements:
                if element.__class__.__name__ == 'Spring':  # FIXME:
                    continue
                element_data = {}
                item = Item(element_data, element.id)
                element.draw(item)
                elements_data.append(element_data)

            timestep_data = {
                'nodes': nodes_data,
                'elements': elements_data,
            }

            timesteps_data.append(timestep_data)

        data = {
            'timesteps': timesteps_data
        }

        template_path = os.path.join(os.path.dirname(__file__), 'html', 'index.html')
        with open(template_path, 'r', encoding='UTF-8') as file:
            template = file.read()

        template = self._embed_js(template, 'index.js')

        content = template.replace("const data = {}", "const data = " + json.dumps(data))
        
        return content

    def show(self, height, model):
        display_html(self.html(height, model), raw=True)



"""Strategies for solving linear and nonlinear FE problems."""

from __future__ import annotations

import nfem
from nfem.assembler import Assembler
from nfem.element import Element
from nfem.model_status import ModelStatus
from nfem.viewer import load_html, show_html

import numpy as np
import numpy.linalg as la

import io
from typing import Tuple, List


def solve_linear(model: nfem.Model):
    """Solve the linear FE problem."""
    assembler = Assembler(model)

    load_factor_hat = model.load_factor

    n, m = assembler.size

    r = np.empty(m)
    k = np.empty((m, m))

    rhs = np.empty(n + 1)
    lhs = np.empty((n + 1, n + 1))

    rnorm_history: List[float] = []
    dnorm_history: List[float] = []

    load_factor_hat = model.load_factor
    tolerance = 1e-6
    max_iterations = 2

    iteration = 0

    while True:
        # compute residual forces of the system

        for i in range(m):
            r[i] = -assembler.dofs[i].external_force

        r *= model.load_factor

        assembler.assemble_vector(_element_linear_r, out=r)

        # build right-hand-side of the equation system

        rhs[:n] = -r[:n]
        rhs[n] = model.load_factor - load_factor_hat

        # check residual criterion

        rnorm = la.norm(rhs)
        rnorm_history.append(rnorm)

        if rnorm < tolerance:
            break

        # check iteration criterion

        if iteration > max_iterations:
            break

        # create a duplicate of the current state before updating and insert it
        # in the history

        duplicate = model.get_duplicate()
        duplicate._previous_model = model._previous_model
        model._previous_model = duplicate
        duplicate.status = model.status
        model.status = ModelStatus.iteration

        # compute stiffness matrix of the system

        k.fill(0)
        assembler.assemble_matrix(_element_linear_k, out=k)

        # build left-hand-side of the equation system

        lhs.fill(0)
        lhs[:n, :n] = k[:n, :n]
        lhs[n, n] = 1
        for i in range(n):
            lhs[i, n] = -assembler.dofs[i].external_force

        # solve linear equation system: lhs * delta = rhs

        try:
            delta = la.solve(lhs, rhs)
        except np.linalg.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        dnorm = la.norm(delta)
        dnorm_history.append(dnorm)

        # update model

        for i in range(n):
            assembler.dofs[i].value += delta[i]

        model.load_factor += delta[n]

        iteration += 1

    if iteration > max_iterations:
        raise RuntimeError(
            f'Newthon-Raphson did not converge after {max_iterations} steps.' +
            f' Residual norm: {rnorm}'
        )

    # update residual forces

    for i in range(m):
        assembler.dofs[i].residual = r[i]

    model.status = ModelStatus.equilibrium

    return SolutionInfo(True, rnorm_history, dnorm_history)


def solve_load_control(model: nfem.Model, tolerance: float = 1e-5,
                       max_iterations: int = 100):
    """Solve the nonlinear FE problem using load control."""
    assembler = Assembler(model)

    load_factor_hat = model.load_factor

    n, m = assembler.size

    r = np.empty(m)
    k = np.empty((m, m))

    rhs = np.zeros(n + 1)
    lhs = np.zeros((n + 1, n + 1))

    rnorm_history: List[float] = []
    dnorm_history: List[float] = []

    iteration = 0

    while True:
        # compute residual forces of the system

        for i in range(m):
            r[i] = -assembler.dofs[i].external_force

        r *= model.load_factor

        assembler.assemble_vector(_element_r, out=r)

        # build right-hand-side of the equation system

        rhs[:n] = -r[:n]
        rhs[n] = model.load_factor - load_factor_hat

        # check residual criterion

        rnorm = la.norm(rhs)
        rnorm_history.append(rnorm)

        if rnorm < tolerance:
            break

        # check iteration criterion

        if iteration > max_iterations:
            break

        # create a duplicate of the current state before updating and insert it
        # in the history

        duplicate = model.get_duplicate()
        duplicate._previous_model = model._previous_model
        model._previous_model = duplicate
        duplicate.status = model.status
        model.status = ModelStatus.iteration

        # compute stiffness matrix of the system

        k.fill(0)
        assembler.assemble_matrix(_element_k, out=k)

        # build left-hand-side of the equation system

        lhs.fill(0)
        lhs[:n, :n] = k[:n, :n]
        lhs[n, n] = 1
        for i in range(n):
            lhs[i, n] = -assembler.dofs[i].external_force

        # solve linear equation system: lhs * delta = rhs

        try:
            delta = la.solve(lhs, rhs)
        except np.linalg.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        dnorm = la.norm(delta)
        dnorm_history.append(dnorm)

        # update model

        for i in range(n):
            assembler.dofs[i].value += delta[i]

        model.load_factor += delta[n]

        iteration += 1

    if iteration > max_iterations:
        raise RuntimeError(
            f'Newthon-Raphson did not converge after {max_iterations} steps.' +
            f' Residual norm: {rnorm}'
        )

    # update residual forces

    for i in range(m):
        assembler.dofs[i].residual = r[i]

    model.status = ModelStatus.equilibrium

    return SolutionInfo(True, rnorm_history, dnorm_history)


def solve_displacement_control(model: nfem.Model, dof: Tuple[str, str],
                               tolerance: float = 1e-5,
                               max_iterations: int = 100):
    """Solve the nonlinear FE problem using displacement control."""
    assembler = Assembler(model)

    dof_index = assembler.dof_indices[dof]
    d_hat = assembler.dofs[dof_index].delta

    n, m = assembler.size

    r = np.empty(m)
    k = np.empty((m, m))

    rhs = np.zeros(n + 1)
    lhs = np.zeros((n + 1, n + 1))

    rnorm_history: List[float] = []
    dnorm_history: List[float] = []

    iteration = 0

    while True:
        # compute residual forces of the system

        for i in range(m):
            r[i] = -assembler.dofs[i].external_force
        r *= model.load_factor

        assembler.assemble_vector(_element_r, out=r)

        # build right-hand-side of the equation system

        rhs.fill(0)
        rhs[:n] = -r[:n]
        rhs[n] = model[dof].delta - d_hat

        # check residual criterion

        rnorm = la.norm(rhs)
        rnorm_history.append(rnorm)

        if rnorm < tolerance:
            break

        # check iteration criterion

        if iteration > max_iterations:
            break

        # create a duplicate of the current state before updating and insert it
        # in the history

        duplicate = model.get_duplicate()
        duplicate._previous_model = model._previous_model
        model._previous_model = duplicate
        duplicate.status = model.status
        model.status = ModelStatus.iteration

        # compute stiffness matrix of the system

        k.fill(0)
        assembler.assemble_matrix(_element_k, out=k)

        # build left-hand-side of the equation system

        lhs.fill(0)
        lhs[:n, :n] = k[:n, :n]
        lhs[n, dof_index] = 1
        for i in range(n):
            lhs[i, n] = -assembler.dofs[i].external_force

        # solve linear equation system: lhs * delta = rhs

        try:
            delta = la.solve(lhs, rhs)
        except np.linalg.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        dnorm = la.norm(delta)
        dnorm_history.append(dnorm)

        # update model

        for i in range(n):
            assembler.dofs[i].value += delta[i]

        model.load_factor += delta[n]

        iteration += 1

    if iteration > max_iterations:
        raise RuntimeError(
            f'Newthon-Raphson did not converge after {max_iterations} steps.' +
            f' Residual norm: {rnorm}'
        )

    # update residual forces

    for i in range(m):
        assembler.dofs[i].residual = r[i]

    model.status = ModelStatus.equilibrium

    return SolutionInfo(True, rnorm_history, dnorm_history)


def solve_arc_length_control(model: nfem.Model, tolerance: float = 1e-5,
                             max_iterations: int = 100):
    """Solve the nonlinear FE problem using arc-length control."""
    assembler = Assembler(model)

    previous_model = model.get_previous_model()

    def compute_squared_arc_length(model):
        squared_l = (model.load_factor - previous_model.load_factor)**2

        for node, previous_node in zip(model.nodes, previous_model.nodes):
            d = node.location - previous_node.location
            squared_l += d @ d

        return squared_l

    squared_l_hat = compute_squared_arc_length(model)

    n, m = assembler.size

    r = np.empty(m)
    k = np.empty((m, m))

    rhs = np.zeros(n + 1)
    lhs = np.zeros((n + 1, n + 1))

    rnorm_history: List[float] = []
    dnorm_history: List[float] = []

    iteration = 0

    while True:
        # compute residual forces of the system

        for i in range(m):
            r[i] = -assembler.dofs[i].external_force
        r *= model.load_factor

        assembler.assemble_vector(_element_r, out=r)

        # build right-hand-side of the equation system

        rhs.fill(0)
        rhs[:n] = -r[:n]
        rhs[n] = -compute_squared_arc_length(model) + squared_l_hat

        # check residual criterion

        rnorm = la.norm(rhs)
        rnorm_history.append(rnorm)

        if rnorm < tolerance:
            break

        # check iteration criterion

        if iteration > max_iterations:
            break

        # create a duplicate of the current state before updating and insert it
        # in the history

        duplicate = model.get_duplicate()
        duplicate._previous_model = model._previous_model
        model._previous_model = duplicate
        duplicate.status = model.status
        model.status = ModelStatus.iteration

        # compute stiffness matrix of the system

        k.fill(0)
        assembler.assemble_matrix(_element_k, out=k)

        # build left-hand-side of the equation system

        lhs.fill(0)
        lhs[:n, :n] = k[:n, :n]
        for i in range(n):
            lhs[i, n] = -assembler.dofs[i].external_force

        for i, dof in enumerate(assembler.dofs[:n]):
            lhs[n, i] = 2 * (model[dof].delta - previous_model[dof].delta)

        lhs[n, n] = 2 * (model.load_factor - previous_model.load_factor)

        # solve linear equation system: lhs * delta = rhs

        try:
            delta = la.solve(lhs, rhs)
        except np.linalg.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        dnorm = la.norm(delta)
        dnorm_history.append(dnorm)

        # update model

        for i in range(n):
            assembler.dofs[i].value += delta[i]

        model.load_factor += delta[n]

        iteration += 1

    if iteration > max_iterations:
        raise RuntimeError(
            f'Newthon-Raphson did not converge after {max_iterations} steps.' +
            f' Residual norm: {rnorm}'
        )

    # update residual forces

    for i in range(m):
        assembler.dofs[i].residual = r[i]

    model.status = ModelStatus.equilibrium

    solution_info = SolutionInfo(True, rnorm_history, dnorm_history)
    solution_info.message = f'Arc-Length-Control with length = {squared_l_hat**0.5}'

    return solution_info


def _element_linear_r(element: Element):
    return element.compute_linear_r()


def _element_linear_k(element: Element):
    return element.compute_linear_k()


def _element_r(element: Element):
    return element.compute_r()


def _element_k(element: Element):
    return element.compute_k()


class SolutionInfo:
    """Container for linear solution info."""

    message: str

    def __init__(self, converged: bool, rnorm: List[float], dnorm: List[float]):
        """Create a new linear SolutionInfo."""
        self.converged = converged
        self.rnorm = rnorm
        self.dnorm = dnorm

    def __repr__(self) -> str:
        """Get a text representation of the object."""
        output = io.StringIO()
        if self.converged:
            print('System converged!', file=output)
        else:
            print('System not converged!', file=output)
        print(f'# Iterations  = {len(self.rnorm) - 1}', file=output)
        print(f'Residual Norm = {self.rnorm[-1]}', file=output)
        contents = output.getvalue()
        output.close()
        return contents

    def html(self) -> str:
        data = dict(
            converged=self.converged,
            rnorm=self.rnorm,
            xnorm=self.dnorm,
        )

        return load_html('solution-viewer', data)

    def _repr_html_(self) -> str:
        return self.html()

    def show(self) -> None:
        raw_html = self.html()

        show_html(raw_html)



"""Model of a nonlinear finite element problem."""

from __future__ import annotations

from nfem import solve
from nfem.assembler import Assembler
from nfem.dof import Dof
from nfem.element import Element
from nfem.key_collection import KeyCollection
from nfem.model_status import ModelStatus
from nfem.node import Node
from nfem.spring import Spring
from nfem.truss import Truss
from nfem.viewer import show_html

import numpy as np
import numpy.linalg as la
import numpy.typing as npt

from scipy.linalg import eig

from copy import deepcopy
from typing import List, Optional, Sequence, Tuple, Type, Union

DofID = Union[str, Dof, Tuple[str, str]]

Vector = npt.NDArray[np.float64]
Matrix = npt.NDArray[np.float64]


class Model:
    """Model of a nonlinear finite element problem."""

    def __init__(self, name: str = None):
        """Create a new model.

        :name: Name of the model.
        """
        self.name: str = name
        self.status: ModelStatus = ModelStatus.initial
        self.nodes: KeyCollection[str, Node] = KeyCollection()
        self.elements: KeyCollection[str, Element] = KeyCollection()
        self.load_factor: float = 0.0
        self._previous_model: Model = None
        self.det_k: Optional[float] = None
        self.first_eigenvalue: Optional[float] = None
        self.first_eigenvector_model: Optional[Vector] = None

    # === modeling

    def add_node(self, id: str, x: float, y: float, z: float,
                 support: str = '', fx: float = 0.0, fy: float = 0.0,
                 fz: float = 0.0) -> None:
        """Add a three dimensional node to the model.

        Parameters
        ----------
        @id: Unique ID of the node.
        @x: X coordinate.
        @y: Y coordinate.
        @z: Z coordinate.
        @support: Directions with supports.
        @fx: External force in x direction.
        @fy: External force in y direction.
        @fz: External force in z direction.
        """
        if not isinstance(id, str):
            raise TypeError('The node id is not a text string')

        if id in self.nodes:
            raise KeyError(f'The model already contains a node with id {id}')

        node = Node(id, x, y, 0 if z is None else z)

        self.nodes._add(node)

        if 'x' in support:
            node.dof('u').is_active = False
        if 'y' in support:
            node.dof('v').is_active = False
        if 'z' in support:
            node.dof('w').is_active = False

        node.dof('u').external_force = fx
        node.dof('v').external_force = fy
        node.dof('w').external_force = fz

    def add_truss(self, id: str, node_a: str, node_b: str,
                  youngs_modulus: float, area: float, prestress: float = 0.0,
                  tensile_strength: Optional[float] = None,
                  compressive_strength: Optional[float] = None) -> None:
        """Add a three dimensional truss element to the model.

        @id: Unique ID of the element.
        @node_a: ID of the first node.
        @node_b: ID of the second node.
        @youngs_modulus: Youngs modulus of the material for the truss.
        @area: Area of the cross section for the truss.
        @tensile_strength: Tensile strength of the truss.
        @compressive_strength: Compressive strength of the truss.

        Examples
        --------
        Add a truss element from node `A` to node `B`:

        >>> model.add_truss(node_a='A', node_a='B', youngs_modulus=20, area=1)
        """
        if not isinstance(id, str):
            raise TypeError('The element id is not a text string')

        if not isinstance(node_a, str):
            raise TypeError('The id of node_a is not a text string')

        if not isinstance(node_b, str):
            raise TypeError('The id of node_b is not a text string')

        if id in self.elements:
            raise KeyError('The model already contains an element with id ' +
                           f'"{id}"')

        if node_a not in self.nodes:
            raise KeyError('The model does not contain a node with id ' +
                           f'"{node_a}"')

        if node_b not in self.nodes:
            raise KeyError('The model does not contain a node with id ' +
                           f'"{node_b}"')

        element = Truss(id, self.nodes[node_a], self.nodes[node_b],
                        youngs_modulus, area, prestress)
        element.tensile_strength = tensile_strength
        element.compressive_strength = compressive_strength

        self.elements._add(element)

    def add_spring(self, id: str, node: str, kx: float = 0.0, ky: float = 0.0,
                   kz: float = 0.0) -> None:
        """Add a spring element.

        @id: Unique ID.
        @node: Adjacent node.
        @kx: Stiffness in x direction.
        @ky: Stiffness in y direction.
        @kz: Stiffness in z direction.
        """
        if not isinstance(id, str):
            raise TypeError('The element id is not a text string')

        if not isinstance(node, str):
            raise TypeError('The id of node is not a text string')

        if id in self.elements:
            raise KeyError('The model already contains an element with id ' +
                           f'"{id}"')

        if node not in self.nodes:
            raise KeyError('The model does not contain a node with id ' +
                           f'"{node}"')

        element = Spring(id, self.nodes[node], kx, ky, kz)

        self.elements._add(element)

    def add_element(self, element_type: Type, id: str, nodes: List[Node],
                    *args, **kwargs) -> None:
        """Add an element (Advanced!).

        @element_type: Factory for the new element.
        @id: Unique ID.
        @nodes: Adjacent Nodes.
        @args: Additional arguments passed to the factory.
        @kwargs: Additional arguments passed to the factory.
        """
        if not isinstance(id, str):
            raise TypeError('The element id is not a text string')

        if id in self.elements:
            raise KeyError('The model already contains an element with id ' +
                           f'"{id}"')

        node_list = []

        for node in nodes:
            if not isinstance(node, str):
                raise TypeError(f'The id "{node}" is not a text string')

            if node not in self.nodes:
                raise KeyError('The model does not contain a node with id ' +
                               f'"{node}"')

            node_list.append(self.nodes[node])

        element = element_type(id, node_list, *args, **kwargs)

        self.elements._add(element)

    # === degree of freedoms

    def __getitem__(self, key) -> Dof:
        """Get a degree of freedom."""
        if isinstance(key, str):
            return self.dof(key)
        elif isinstance(key, Dof):
            node_key, dof_type = key.id
        else:
            node_key, dof_type = key
        return self.nodes[node_key].dof(dof_type)

    @property
    def dofs(self) -> Sequence[Dof]:
        """Get all degrees of freedom."""
        assembler = Assembler(self)

        n, _ = assembler.size

        return assembler.dofs[:n]

    def dof(self, id: str) -> Dof:
        """Get a degree of freedom."""
        node, direction = id.rsplit('.', maxsplit=1)
        return self.nodes[node].dof(direction)

    # === r and k

    def compute_linear_r(self) -> Vector:
        """Compute the linear residual force vector of the element."""
        assembler = Assembler(self)

        n, m = assembler.size

        r = np.empty(m)

        for i, dof in enumerate(assembler.dofs):
            r[i] = -dof.external_force

        r[:] *= self.load_factor

        def compute_local(element: Element) -> Vector:
            return element.compute_linear_r()

        r = assembler.assemble_vector(compute_local)

        return r[:n]

    def compute_linear_k(self) -> Matrix:
        """Compute the linear stiffness matrix of the element."""
        assembler = Assembler(self)

        n, _ = assembler.size

        def compute_local(element: Element) -> Vector:
            return element.compute_linear_k()

        k = assembler.assemble_vector(compute_local)

        return k[:n, :n]

    def compute_linear_kg(self) -> Matrix:
        """Compute the linear geometric stiffness matrix of the element."""
        assembler = Assembler(self)

        n, _ = assembler.size

        def compute_local(element: Element) -> Vector:
            return element.compute_linear_kg()

        k = assembler.assemble_matrix(compute_local)

        return k[:n, :n]

    def compute_r(self) -> Vector:
        """Compute the nonlinear residual force vector of the element."""
        assembler = Assembler(self)

        n, m = assembler.size

        r = np.empty(m)

        for i, dof in enumerate(assembler.dofs):
            r[i] = -dof.external_force

        r[:] *= self.load_factor

        def compute_local(element: Element) -> Vector:
            return element.compute_r()

        assembler.assemble_vector(compute_local, out=r)

        return r[:n]

    def compute_k(self) -> Matrix:
        """Compute the nonlinear stiffness matrix of the element."""
        assembler = Assembler(self)

        n, _ = assembler.size

        def compute_local(element: Element) -> Vector:
            return element.compute_k()

        k = assembler.assemble_matrix(compute_local)

        return k[:n, :n]

    def compute_ke(self) -> Matrix:
        """Compute the elastic stiffness matrix of the element."""
        assembler = Assembler(self)

        n, _ = assembler.size

        def compute_local(element: Element) -> Vector:
            return element.compute_ke()

        k = assembler.assemble_matrix(compute_local)

        return k[:n, :n]

    def compute_km(self) -> Matrix:
        """Compute the material stiffness matrix of the element."""
        assembler = Assembler(self)

        n, _ = assembler.size

        def compute_local(element: Element) -> Vector:
            return element.compute_km()

        k = assembler.assemble_matrix(compute_local)

        return k[:n, :n]

    def compute_kg(self) -> Matrix:
        """Compute the geometric stiffness matrix of the element."""
        assembler = Assembler(self)

        n, _ = assembler.size

        def compute_local(element: Element) -> Vector:
            return element.compute_kg()

        k = assembler.assemble_matrix(compute_local)

        return k[:n, :n]

    def compute_kd(self) -> Matrix:
        """Compute the initial-displacement stiffness matrix of the element."""
        assembler = Assembler(self)

        n, _ = assembler.size

        def compute_local(element: Element) -> Vector:
            return element.compute_kd()

        k = assembler.assemble_matrix(compute_local)

        return k[:n, :n]

    def compute_det_k(self) -> float:
        """Compute the determinant of k."""
        k = self.compute_k()

        return la.det(k)

    # === increment

    def get_dof_increment(self, dof) -> float:
        """Get the increment of the dof during the last solution step.

        @dof: Dof that is asked.
        """
        if self.get_previous_model() is None:
            return 0.0

        previous_model = self.get_previous_model()

        current_value = self[dof].delta
        previous_value = previous_model[dof].delta

        return current_value - previous_value

    def get_lam_increment(self) -> float:
        """Get the increment of lambda during the last solution step."""
        if self.get_previous_model() is None:
            return 0.0

        current_value = self.load_factor
        previous_value = self.get_previous_model().load_factor

        return current_value - previous_value

    def get_increment_vector(self, assembler: Assembler = None) -> Vector:
        """Get the increment that resulted in the current position.

        @assembler: Initialized assembler.
        """
        if assembler is None:
            assembler = Assembler(self)

        n, _ = assembler.size

        increment = np.zeros(n + 1)

        if self.get_previous_model() is None:
            print('WARNING: Increment is zero because no previous model' +
                  ' exists!')
            return increment

        for index, dof in enumerate(assembler.dofs[:n]):
            increment[index] = self.get_dof_increment(dof)

        increment[-1] = self.get_lam_increment()

        return increment

    def get_increment_norm(self, assembler: Assembler = None) -> float:
        """Get the vector norm of the current increment.

        @assembler: Initialized assembler.
        """
        increment = self.get_increment_vector(assembler)
        return la.norm(increment)

    # === model history

    def get_previous_model(self, skip_iterations: bool = True) -> Model:
        """Get the previous model of the current model.

        @skip_iterations: True if iterations or predictions should be skipped.
        """
        if not skip_iterations:
            return self._previous_model

        # find the most previous model that is not an iteration or prediction
        previous_model = self._previous_model

        while (previous_model is not None and
               previous_model.status in [ModelStatus.duplicate,
                                         ModelStatus.prediction,
                                         ModelStatus.iteration]):
            previous_model = previous_model._previous_model

        return previous_model

    def get_initial_model(self) -> Model:
        """Get the initial model of this model."""
        current_model = self

        while current_model.get_previous_model() is not None:
            current_model = current_model.get_previous_model()

        return current_model

    def get_model_history(self, skip_iterations: bool = True) -> List[Model]:
        """Get a list of all previous models of this model.

        @skip_iterations: True if non converged models should be considered.
        """
        history = [self]

        current_model = self

        while current_model.get_previous_model(skip_iterations) is not None:
            current_model = current_model.get_previous_model(skip_iterations)

            history = [current_model] + history

        return history

    def get_duplicate(self, name: str = None, branch: bool = False) -> Model:
        r"""Get a duplicate of the model.

        @name: Name of the new model.
        @branch: True to create a new branch.


        Notes
        -----
        If `branch` is `False` the duplicate will be a successor of the
        current model:

            previous ----> current ----> duplicate

        If `branch` is `True` the duplicate will be a successor of the
        previous model:

            previous ----> current
                    \
                     \---> duplicate
        """
        temp_previous_model = self._previous_model
        self._previous_model = None

        duplicate = deepcopy(self)

        self._previous_model = temp_previous_model

        if branch:
            duplicate._previous_model = self._previous_model
        else:
            duplicate._previous_model = self
            duplicate.status = ModelStatus.duplicate

        if name is not None:
            duplicate.name = name

        # make sure the duplicated model is in a clean state
        duplicate.det_k = None
        duplicate.first_eigenvalue = None
        duplicate.first_eigenvector_model = None

        return duplicate

    def new_timestep(self, name: str = None) -> Model:
        """Create a new timestep."""
        temp_previous_model = self._previous_model
        self._previous_model = None

        duplicate = deepcopy(self)

        self._previous_model = temp_previous_model

        duplicate._previous_model = self

        if name is not None:
            duplicate.name = name

        duplicate.det_k = None
        duplicate.first_eigenvalue = None
        duplicate.first_eigenvector_model = None

        return duplicate

    # === solving

    def solve_linear(self) -> solve.SolutionInfo:
        """Perform a linear solution step on the model."""
        return solve.solve_linear(self)

    def solve_load_control(self, tolerance: float = 1e-5, max_iterations: int = 500) -> solve.SolutionInfo:
        """Perform a solution step using load control."""
        return solve.solve_load_control(self, tolerance, max_iterations)

    def solve_displacement_control(self, dof: DofID, tolerance: float = 1e-5, max_iterations: int = 500) -> solve.SolutionInfo:
        """Perform a solution step using displacement control."""
        return solve.solve_displacement_control(self, dof, tolerance, max_iterations)

    def solve_arc_length_control(self, tolerance: float = 1e-5, max_iterations: int = 500) -> solve.SolutionInfo:
        """Perform a solution step using arc-length control."""
        return solve.solve_arc_length_control(self, tolerance, max_iterations)

    # ---

    def perform_linear_solution_step(self, info: bool = False) -> None:
        """Perform a linear solution step on the model.

        It uses the current load factor.
        The results are stored at the dofs and used to update the current
        coordinates of the nodes.
        """
        solution = self.solve_linear()

        if info:
            print(f'Linear solution with λ = {self.load_factor}')
            solution.show()
            print()

    def perform_load_control_step(self,
                                  tolerance: float = 1e-5,
                                  max_iterations: int = 500,
                                  info: bool = False,
                                  solve_det_k: bool = True,
                                  solve_attendant_eigenvalue: bool = False
                                  ) -> None:
        """Perform a solution step using load control."""
        solution = solve.solve_load_control(self, tolerance, max_iterations)

        if solve_det_k or solve_attendant_eigenvalue:
            assembler = Assembler(self)

            n, m = assembler.size

        if solve_det_k:
            k = np.zeros((m, m))
            assembler.assemble_matrix(lambda e: e.compute_k(), out=k)
            self.det_k = la.det(k[:n, :n])

        if solve_attendant_eigenvalue:
            self.solve_eigenvalues(assembler=assembler)

        if info:
            print(f'Load-Control with λ = {self.load_factor}')
            solution.show()
            print()

    def perform_displacement_control_step(self,
                                          dof: DofID,
                                          tolerance: float = 1e-5,
                                          max_iterations: int = 500,
                                          info: bool = False,
                                          solve_det_k: bool = True,
                                          solve_attendant_eigenvalue: bool
                                          = False
                                          ) -> None:
        """Perform a solution step using displacement control."""
        solution = self.solve_displacement_control(dof, tolerance,
                                                   max_iterations)

        if solve_det_k or solve_attendant_eigenvalue:
            assembler = Assembler(self)

            n, m = assembler.size

        if solve_det_k:
            k = np.zeros((m, m))
            assembler.assemble_matrix(lambda e: e.compute_k(), out=k)
            self.det_k = la.det(k[:n, :n])

        if solve_attendant_eigenvalue:
            self.solve_eigenvalues(assembler=assembler)

        if info:
            print(f'Displacement-Control with {dof[1]} at node {dof[0]} =' +
                  f' {self[dof].delta}')
            solution.show()
            print()

    def perform_arc_length_control_step(self,
                                        tolerance: float = 1e-5,
                                        max_iterations: int = 500,
                                        info: bool = False,
                                        solve_det_k: bool = True,
                                        solve_attendant_eigenvalue: bool
                                        = False
                                        ) -> None:
        """Perform a solution step using arc-length control."""
        solution_info = self.solve_arc_length_control(tolerance, max_iterations)

        if solve_det_k or solve_attendant_eigenvalue:
            assembler = Assembler(self)

            n, m = assembler.size

        if solve_det_k:
            k = np.zeros((m, m))
            assembler.assemble_matrix(lambda e: e.compute_k(), out=k)
            self.det_k = la.det(k[:n, :n])

        if solve_attendant_eigenvalue:
            self.solve_eigenvalues(assembler=assembler)

        if info:
            print(solution_info.message)
            solution_info.show()
            print()

    def solve_det_k(self) -> None:
        """Compute, stores and prints the determinant of k."""
        self.det_k = self.compute_det_k()

        print(f'Det(K): {self.det_k}')

    def solve_linear_eigenvalues(self, assembler: Assembler = None) -> None:
        """Solve the linearized eigenvalue problem.

        @assembler: Initialized assembler.

        Eigenvalue problem:
        [ k_e + eigvals * k_g(linear strain) ] * eigvecs = 0

        Stores the first positive eigenvalue and vector
        """
        if assembler is None:
            assembler = Assembler(self)

        n, m = assembler.size

        # assemble matrices
        k_e = np.zeros((m, m))
        k_g = np.zeros((m, m))
        print("=================================")
        print('Linearized prebuckling (LPB) analysis ...')
        assembler.assemble_matrix(
            lambda element: element.compute_linear_k(), out=k_e)
        assembler.assemble_matrix(
            lambda element: element.compute_linear_kg(), out=k_g)

        # solve eigenvalue problem
        eigvals, eigvecs = eig(k_e[:n, :n], -k_g[:n, :n])

        # extract real parts of eigenvalues
        eigvals = np.array([x.real for x in eigvals])

        # sort eigenvalues and vectors
        idx = eigvals.argsort()
        eigvals = eigvals[idx]
        eigvecs = eigvecs[:, idx]

        # remove negative eigenvalues
        for i, eigenvalue in enumerate(eigvals):
            if eigenvalue > 0.0:
                break
            else:
                i += 1

        if i == len(eigvals):
            print('System has no positive eigenvalues!')
            return

        eigvals = eigvals[i:]
        eigvecs = eigvecs[:, i:]

        print(f'First linear eigenvalue: {eigvals[0]}')
        # this is printed in TRUSS
        print('First linear eigenvalue * lambda:' +
              f' {eigvals[0] * self.load_factor}')
        if len(eigvecs[0]) < 10:
            print(f'First linear eigenvector: {eigvecs[0]}')

        self.first_eigenvalue = eigvals[0]

        # store eigenvector as model
        model = self.get_duplicate()
        model._previous_model = self
        model.status = ModelStatus.eigenvector
        model.det_k = None
        model.first_eigenvalue = None
        model.first_eigenvector_model = None
        model.load_factor = None

        for index, dof in enumerate(assembler.dofs[:n]):
            model[dof].delta = eigvecs[index][0]

        self.first_eigenvector_model = model

    def solve_eigenvalues(self, assembler: Assembler = None) -> None:
        """Solve the eigenvalue problem.

        @assembler: Initialized assembler.

        Eigenvalue problem:
        [ k_m + eigvals * k_g ] * eigvecs = 0

        Stores the closest (most critical) eigenvalue and vector
        """
        if assembler is None:
            assembler = Assembler(self)

        n, m = assembler.size

        # assemble matrices
        k_m = np.zeros((m, m))
        k_g = np.zeros((m, m))
        print("=================================")
        print('Attendant eigenvalue analysis ...')
        assembler.assemble_matrix(
            lambda element: element.compute_km(), out=k_m)
        assembler.assemble_matrix(
            lambda element: element.compute_kg(), out=k_g)

        # solve eigenvalue problem
        eigvals, eigvecs = eig(k_m[:n, :n], -k_g[:n, :n])

        # extract real parts of eigenvalues
        eigvals = np.array([x.real for x in eigvals])

        # sort eigenvalues and vectors
        idx = eigvals.argsort()
        eigvals = eigvals[idx]
        eigvecs = eigvecs[:, idx]

        # find index of closest eigenvalue to 1 (we could store all but that
        # seems like an overkill)
        idx = (np.abs(eigvals - 1.0)).argmin()

        print(f'Closest eigenvalue: {eigvals[idx]}')
        # this is printed in TRUSS
        print(
            f'Closest eigenvalue * lambda: {eigvals[idx] * self.load_factor}')
        if len(eigvecs[idx]) < 10:
            print('Closest eigenvector: {eigvecs[idx]}')

        self.first_eigenvalue = eigvals[idx]

        # store eigenvector as model
        model = self.get_duplicate()
        model._previous_model = self
        model.status = ModelStatus.eigenvector
        model.det_k = None
        model.first_eigenvalue = None
        model.first_eigenvector_model = None
        model.load_factor = None

        for index, dof in enumerate(assembler.dofs[:n]):
            model[dof].delta = eigvecs[index][idx]

        self.first_eigenvector_model = model

    def get_tangent_vector(self, assembler: Assembler = None) -> Vector:
        """Get the tangent vector.

        @tangent: Tangent vector t = [v, 1]
                  with v = d_u / d_lambda ... incremental velocity
        """
        if assembler is None:
            assembler = Assembler(self)

        n, m = assembler.size

        tangent = np.zeros(n + 1)

        k = np.zeros((m, m))
        external_f = np.zeros(n)

        # assemble stiffness
        assembler.assemble_matrix(lambda element: element.compute_k(), out=k)

        # assemble force

        for i in range(n):
            external_f[i] += assembler.dofs[i].external_force

        try:
            tangent[:n] = la.solve(k[:n, :n], external_f)
        except la.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        # load-factor = 1

        tangent[n] = 1

        return tangent

    # === prediction functions

    def predict_load_factor(self, value: float) -> None:
        """Predict the solution by predictor_method lambda.

        @value: Value for the new load factor lambda.
        """
        self.status = ModelStatus.prediction
        self.load_factor = value

    def predict_load_increment(self, value: float) -> None:
        """Predict the solution by incrementing lambda.

        @value: Value that is used to increment the load factor lambda.
        """
        self.status = ModelStatus.prediction
        self.load_factor += value

    def predict_dof_state(self, dof: DofID, value: float) -> None:
        """Predict the solution by predictor_method the dof.

        @dof: Dof that is prescribed.
        @value: Value that is used to prescribe the dof.
        """
        self.status = ModelStatus.prediction
        self[dof].delta = value

    def predict_dof_increment(self, dof: DofID, value: float) -> None:
        """Predict the solution by incrementing the dof.

        @dof: Dof that is incremented.
        @value: Value that is used to increment the dof.
        """
        self.status = ModelStatus.prediction
        self[dof].delta += value

    def predict_with_last_increment(self,
                                    value: Optional[float] = None
                                    ) -> None:
        """Predict the solution by incrementing lambda and all dofs.

        Uses the increment of the last solution step.

        @value: Length of the increment
        """
        self.status = ModelStatus.prediction
        if self.get_previous_model().get_previous_model() is None:
            raise RuntimeError('predict_with_last_increment can only be used' +
                               ' after the first step!')

        assembler = Assembler(self)

        previous_model = self.get_previous_model()

        last_increment = previous_model.get_increment_vector(assembler)

        length = la.norm(last_increment)

        if value is not None and length != 0.0:
            last_increment *= value / length
            length *= value

        if length == 0.0:
            print("WARNING: The length of the prescribed increment is 0.0!")

        # update dofs at model
        for index, dof in enumerate(assembler.dofs):
            self[dof].delta += last_increment[index]

        # update lam at model
        self.load_factor += last_increment[-1]

    def predict_tangential(self, strategy, **options):
        """Make a tangential prediction.

        Predicts the solution by incrementing lambda and all dofs with the
        increment of the last solution step.

        Parameters
        ----------
        strategy : str
            Strategy to scale the tangent vector.
            Avaiblable options:
            - 'lambda'
            - 'delta-lambda'
            - 'dof'
            - 'delta-dof'
            - 'arc-length'
        options
            value : float
                prescribed value according to the strategy
            dof : Object
                specifies the controlled dof for 'dof' and 'delta-dof' strategy
        """
        self.status = ModelStatus.prediction
        assembler = Assembler(self)

        n, _ = assembler.size

        # get tangent vector
        tangent = self.get_tangent_vector(assembler=assembler)

        # calculate scaling factor according to chosen strategy
        if strategy == 'lambda':
            prescribed_lam = options['value']

            delta_lambda = prescribed_lam - self.load_factor
            factor = delta_lambda

        elif strategy == 'delta-lambda':
            delta_lambda = options['value']

            factor = delta_lambda

        elif strategy == 'dof':
            dof = options['dof']
            value_prescribed = options['value']

            value = self[dof].delta

            delta_prescribed = value_prescribed - value

            dof_index = assembler.dof_indices[dof]
            delta_current = tangent[dof_index]

            factor = delta_prescribed / delta_current

        elif strategy == 'delta-dof':
            dof = options['dof']
            delta_dof_prescribed = options['value']

            dof_index = assembler.dof_indices[dof]
            delta_dof_current = tangent[dof_index]

            factor = delta_dof_prescribed / delta_dof_current

        elif strategy == 'arc-length':
            previous_model = self.get_previous_model()

            if previous_model.get_previous_model() is not None:
                previous_increment = previous_model.get_increment_vector(
                    assembler)

            if 'value' in options.keys():
                prescribed_length = options['value']
            elif previous_model.get_previous_model() is not None:
                prescribed_length = la.norm(previous_increment)
            else:
                prescribed_length = 0.0

            if prescribed_length == 0.0:
                print("WARNING: Length of the prescribed increment is 0.0!")

            current_length = la.norm(tangent)

            factor = prescribed_length / current_length

            # tangent should point in a similar direction as the last increment
            if previous_model.get_previous_model() is not None:
                if previous_increment @ tangent < 0:
                    factor = -factor

        else:
            raise ValueError(f'Invalid strategy for prediction: {strategy}')

        # scale tangent vector
        tangent *= factor

        # update dofs at model
        for i, dof in enumerate(assembler.dofs[:n]):
            self[dof].delta += tangent[i]

        # update lambda at model
        self.load_factor += tangent[-1]

    def combine_prediction_with_eigenvector(self, beta):
        """Combine the prediciton with the first eigenvector.

        @beta: Factor between -1.0 and 1.0 used for a linear combination of the
               prediction with the eigenvector.

        :raises RuntimeError: If the model is not in prediction status.
        :raises ValueError: If the beta is not between -1.0 and 1.0.
        """
        if self.status != ModelStatus.prediction:
            raise RuntimeError('Model is not a predictor. Cannot combine' +
                               ' with eigenvector!')

        if beta < -1.0 or beta > 1.0:
            raise ValueError('beta needs to be between -1.0 and 1.0')

        previous_model = self.get_previous_model()
        if previous_model.first_eigenvector_model is None:
            print('WARNING: solving eigenvalue problem in order to do branch' +
                  ' switching')
            previous_model.solve_eigenvalues()

        eigenvector_model = previous_model.first_eigenvector_model

        assembler = Assembler(self)

        u_prediction = self.get_delta_dof_vector(previous_model, assembler)

        prediction_length = la.norm(u_prediction)

        eigenvector = eigenvector_model.get_delta_dof_vector(
            assembler=assembler)

        # scale eigenvector to the length of the prediction
        eigenvector *= (1.0/(la.norm(eigenvector)/prediction_length))

        prediction = u_prediction * (1.0 - abs(beta)) + eigenvector * beta

        delta_prediction = prediction - u_prediction

        # lambda = 0 for the eigenvector. Note: TRUSS.xls uses the same value
        # as for the last increment
        delta_lam = - self.get_lam_increment()

        # update dofs at model

        n, _ = assembler.size
        for index, dof in enumerate(assembler.dofs[:n]):
            self[dof].delta += delta_prediction[index]

        # update lambda at model
        self.load_factor += delta_lam

    def scale_prediction(self, factor):
        """Scale the prediction with a factor.

        @factor: Factor used to scale the prediction.

        @raises RuntimeError: If the model is not in prediction status
        @raises RuntimeError: If the model has no previous model
        """
        if self.status != ModelStatus.prediction:
            raise RuntimeError('Model is not a predictor. Can only scale' +
                               ' predictor!')

        if factor == 1.0:
            return

        previous_model = self.get_previous_model()

        if previous_model is None:
            raise RuntimeError('Previous Model is None!')

        assembler = Assembler(self)

        n, _ = assembler.size

        delta_dof_vector = self.get_delta_dof_vector(
            previous_model, assembler=assembler)

        delta_lambda = self.load_factor - previous_model.load_factor

        delta_dof_vector *= (factor - 1.0)
        delta_lambda *= (factor - 1.0)

        for i, dof in enumerate(assembler.dofs[:n]):
            self[dof].delta += delta_dof_vector[i]

        self.load_factor += delta_lambda

    def get_delta_dof_vector(self, model_b: Model = None,
                             assembler: Assembler = None) -> Vector:
        """Get the delta dof between this and a given model_b as a numpy array.

        @model_b: Model that is used as reference for the delta dof
            calculation. If not given, the initial model is used as reference.
        @assembler: Assembler is used to order the dofs in the vector. If not
            given, a new assembler is created.

        @raises RuntimeError: If the model is not in prediction status.
        @raises RuntimeError: If the model has no previous model.
        """
        if model_b is None:
            model_b = self.get_initial_model()

        if assembler is None:
            assembler = Assembler(self)

        n, _ = assembler.size

        delta = np.zeros(n)

        for index, dof in enumerate(assembler.dofs[:n]):
            delta[index] = self[dof].delta - model_b[dof].delta

        return delta

    def load_displacement_curve(self, dof: DofID,
                                skip_iterations: bool = True):
        """Get the load displacement curve for a specific degree of freedom."""
        history = self.get_model_history(skip_iterations)

        data = np.zeros([2, len(history)])

        for i, self in enumerate(history):
            data[0, i] = self[dof].delta
            data[1, i] = self.load_factor

        return data

    def _ipython_display_(self) -> None:
        self.show()

    def html(self, height: int = 500, timestep: int = 0) -> str:
        timesteps = []

        for model in self.get_model_history():
            ts = {}

            ts['name'] = model.name or ''

            ts['objects'] = objects = []

            for element in model.elements:
                element.draw(objects)

            for node in model.nodes:
                node.draw(objects)

            timesteps.append(ts)

        data = dict(
            settings=dict(
                height=height,
                timestep=timestep,
            ),
            timesteps=timesteps,
        )

        from nfem.viewer import load_html

        return load_html('model-viewer', data)

    def show(self, height: int = 500, timestep: int = 0) -> None:
        raw_html = self.html(height, timestep)

        show_html(raw_html, height, iframe=True)



"""Auxiliary for assembling system vectors and system matrices."""

from __future__ import annotations

from nfem.dof import Dof
from nfem.element import Element
import nfem

import numpy as np
import numpy.typing as npt

from typing import Callable, Dict, Sequence, Tuple


EvalElement = Callable[[Element], npt.ArrayLike]


class Assembler:
    """Auxiliary for assembling system vectors and system matrices."""

    def __init__(self, model: nfem.Model):
        """Create a new Assembler.

        :model: Model of the system to be assembled.
        """
        dof_indices: Dict[Dof, int] = {}
        element_indices: Dict[Element, npt.NDArray[int]] = {}

        index = -1

        for element in model.elements:
            indices = np.empty(len(element.dofs), int)

            for i, dof in enumerate(element.dofs):
                dof_index = dof_indices.get(dof, None)

                if dof_index is None:
                    if dof.is_active:
                        index += 1
                        dof_index = index
                    else:
                        dof_index = index - len(dof_indices)

                    dof_indices[dof] = dof_index

                indices[i] = dof_index

            element_indices[element] = indices

        dofs = np.empty(len(dof_indices), object)

        for dof, i in dof_indices.items():
            dofs[i] = dof

        self.dofs: npt.NDArray[Dof] = dofs
        """List of all degrees of freedom including the locked ones."""

        self.dof_indices: Dict[Dof, int] = dof_indices
        """Provides the index for a given degree of freedom."""

        self.element_indices: Dict[Element, npt.NDArray] = element_indices
        """Indices of the degrees of freedom for a given element."""

        self.size: Tuple[int, int] = (index + 1, len(dofs))
        """Number of degrees of freedom (active and total)."""

    def assemble_vector(self, fn: EvalElement, out=None) -> npt.NDArray:
        """Assemble a system vector.

        :fn: Function to compute the local vector of each element.
        :out: Vector to save the result. (optional)
        :return: The assembled system vector.
        """
        if out is None:
            m = len(self.dofs)
            out = np.zeros(m, float)

        for element, indices in self.element_indices.items():
            local_vector = fn(element)
            out[indices] += local_vector

        return out

    def assemble_matrix(self, fn: EvalElement, out=None) -> npt.NDArray:
        """Assemble a system matrix.

        :fn: Function to compute the local matrix of each element.
        :out: matrix to save the result. (optional)
        :return: The assembled system matrix.
        """
        if out is None:
            m = len(self.dofs)
            out = np.zeros((m, m), float)

        for element, indices in self.element_indices.items():
            local_matrix = fn(element)
            out[np.ix_(indices, indices)] += local_matrix

        return out



"""Enueration for the model status."""

from __future__ import annotations

from enum import Enum


class ModelStatus(Enum):
    """Enueration for the model status."""

    initial = 0
    duplicate = 1
    prediction = 2
    iteration = 3
    equilibrium = 4
    eigenvector = 5



from IPython.display import display_html
from xml.dom.minidom import parse as parse_xml
import json
import pathlib
import html


ROOT = pathlib.Path(__file__).parent.resolve()


def load_html(name: str, data):
    path = f'{ROOT}/html/{name}.html'

    with open(path, 'r', encoding='utf-8') as f:
        html = f.read()

    start_tag = '<script type="application/json">'
    end_tag = '</script>'

    start_index = html.find(start_tag) + len(start_tag)

    head = html[:start_index]
    tail = html[start_index:]

    end_index = html.find(end_tag)

    tail = html[end_index:]

    return head + json.dumps(data) + tail


def show_html(raw_html, height=400, iframe=False):
    import sys
    if 'ipykernel' in sys.modules:
        from html import escape
        from IPython.display import display_html

        if iframe:
            display_html(f'<iframe seamless frameborder="0" allowfullscreen width="100%" height="{height}" srcdoc="{escape(raw_html)}"></iframe>', raw=True)
        else:
            display_html(raw_html, raw=True)
    else:
        from PyQt6.QtWebEngineWidgets import QWebEngineView
        from PyQt6.QtWidgets import QApplication

        import sys

        app = QApplication(sys.argv + ['--disable-logging'])

        view = QWebEngineView()
        view.setWindowTitle('nfem')
        view.setHtml(raw_html)
        view.show()

        sys.exit(app.exec())



"""Base type of a finite element."""

from nfem.dof import Dof
from nfem.node import Node

import numpy.typing as npt

from typing import Sequence

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


class Element(Protocol):
    """Base type of a finite element."""

    id: str
    """The unique ID of the element."""

    dofs: Sequence[Dof]
    """A list containing the local degrees of freedom."""

    nodes: Sequence[Node]
    """A list containing the adjacent nodes."""

    def compute_linear_r(self) -> npt.NDArray:
        """Compute the linear residual force vector of the element."""

    def compute_linear_k(self) -> npt.NDArray:
        """Compute the linear stiffness matrix of the element."""

    def compute_linear_kg(self) -> npt.NDArray:
        """Compute the linear geometric stiffness matrix of the element."""

    def compute_r(self) -> npt.NDArray:
        """Compute the nonlinear residual force vector of the element."""

    def compute_k(self) -> npt.NDArray:
        """Compute the nonlinear stiffness matrix of the element."""

    def compute_ke(self) -> npt.NDArray:
        """Compute the elastic stiffness matrix of the element."""

    def compute_km(self) -> npt.NDArray:
        """Compute the material stiffness matrix of the element."""

    def compute_kg(self) -> npt.NDArray:
        """Compute the geometric stiffness matrix of the element."""

    def compute_kd(self) -> npt.NDArray:
        """Compute the initial-displacement stiffness matrix of the element."""



"""Three dimensional Node."""

from __future__ import annotations

from nfem.dof import Dof

import numpy as np
import numpy.typing as npt

from typing import Sequence

Vector = npt.NDArray[np.float64]


class Node:
    """Three dimensional Node.

    :id: Unique ID.
    :ref_x: ref X coordinate.
    :ref_y: ref Y coordinate.
    :ref_z: ref Z coordinate.
    :x: Actual X coordinate.
    :y: Actual Y coordinate.
    :z: Actual Z coordinate.
    :u: Displacement in x direction.
    :v: Displacement in y direction.
    :w: Displacement in z direction.
    """

    id: str
    _dof_x: Dof
    _dof_y: Dof
    _dof_z: Dof

    def __init__(self, id: str, x: float, y: float, z: float):
        """Create a new node.

        :id: Unique ID of the node.
        :x: Initial X coordinate of the node.
        :y: Initial Y coordinate of the node.
        :z: Initial Z coordinate of the node.
        """
        self.id = id
        self._dof_x = Dof(id=(id, 'u'), value=x)
        self._dof_y = Dof(id=(id, 'v'), value=y)
        self._dof_z = Dof(id=(id, 'w'), value=z)

    def dof(self, dof_type: str) -> Dof:
        """Get the degree of freedom for a specific direction."""
        if dof_type == 'u':
            return self._dof_x
        if dof_type == 'v':
            return self._dof_y
        if dof_type == 'w':
            return self._dof_z
        raise AttributeError(f'Node has no dof of type "{dof_type}"')

    # reference location

    @property
    def ref_x(self) -> float:
        """Get or set the x coordinate in the undeformed configuration."""
        return self._dof_x.ref_value

    @ref_x.setter
    def ref_x(self, value: float) -> None:
        self._dof_x.ref_value = value

    @property
    def ref_y(self) -> float:
        """Get or set the y coordinate in the undeformed configuration."""
        return self._dof_y.ref_value

    @ref_y.setter
    def ref_y(self, value: float) -> None:
        self._dof_y.ref_value = value

    @property
    def ref_z(self) -> float:
        """Get or set the z coordinate in the undeformed configuration."""
        return self._dof_z.ref_value

    @ref_z.setter
    def ref_z(self, value: float) -> None:
        self._dof_z.ref_value = value

    @property
    def ref_location(self) -> Vector:
        """Get or set the location in the undeformed configuration."""
        return np.array([self.ref_x, self.ref_y, self.ref_z])

    @ref_location.setter
    def ref_location(self, value: Sequence[float]):
        self.ref_x, self.ref_y, self.ref_z = value

    # actual location

    @property
    def x(self) -> float:
        """Get or set the x coordinate in the deformed configuration."""
        return self._dof_x.value

    @x.setter
    def x(self, value: float) -> None:
        self._dof_x.value = value

    @property
    def y(self) -> float:
        """Get or set the y coordinate in the deformed configuration."""
        return self._dof_y.value

    @y.setter
    def y(self, value: float) -> None:
        self._dof_y.value = value

    @property
    def z(self) -> float:
        """Get or set the z coordinate in the deformed configuration."""
        return self._dof_z.value

    @z.setter
    def z(self, value: float) -> None:
        self._dof_z.value = value

    @property
    def location(self) -> Vector:
        """Get or set the location in the deformed configuration."""
        return np.array([self.x, self.y, self.z])

    @location.setter
    def location(self, value: Sequence[float]):
        self.x, self.y, self.z = value

    # displacements

    @property
    def u(self) -> float:
        """Get or set the displacement in x direction."""
        return self._dof_x.delta

    @u.setter
    def u(self, value: float) -> None:
        self._dof_x.delta = value

    @property
    def v(self) -> float:
        """Get or set the displacement in y direction."""
        return self._dof_y.delta

    @v.setter
    def v(self, value: float) -> None:
        self._dof_y.delta = value

    @property
    def w(self) -> float:
        """Get or set the displacement in z direction."""
        return self._dof_z.delta

    @w.setter
    def w(self, value: float) -> None:
        self._dof_z.delta = value

    @property
    def displacement(self) -> Vector:
        """Get or set the displacement."""
        return np.array([self.u, self.v, self.w])

    @displacement.setter
    def displacement(self, value: Sequence[float]):
        self.u, self.v, self.w = value

    # external forces

    @property
    def fx(self) -> float:
        """Get or set the external force in x direction."""
        return self._dof_x.external_force

    @fx.setter
    def fx(self, value: float) -> None:
        self._dof_x.external_force = value

    @property
    def fy(self) -> float:
        """Get or set the external force in y direction."""
        return self._dof_y.external_force

    @fy.setter
    def fy(self, value: float) -> None:
        self._dof_y.external_force = value

    @property
    def fz(self) -> float:
        """Get or set the external force in z direction."""
        return self._dof_z.external_force

    @fz.setter
    def fz(self, value: float) -> None:
        self._dof_z.external_force = value

    @property
    def external_force(self) -> Vector:
        """Get or set the external force."""
        return np.array([self.fx, self.fy, self.fz])

    @external_force.setter
    def external_force(self, value: Sequence[float]):
        [self.fx, self.fy, self.fz] = value

    # residual forces

    @property
    def rx(self) -> float:
        """Get or set the residual force in x direction."""
        return self._dof_x.residual

    @rx.setter
    def rx(self, value: float) -> None:
        self._dof_x.residual = value

    @property
    def ry(self) -> float:
        """Get or set the residual force in y direction."""
        return self._dof_y.residual

    @ry.setter
    def ry(self, value: float) -> None:
        self._dof_y.residual = value

    @property
    def rz(self) -> float:
        """Get or set the residual force in z direction."""
        return self._dof_z.residual

    @rz.setter
    def rz(self, value: float) -> None:
        self._dof_z.residual = value

    @property
    def residual(self) -> Vector:
        """Get or set the residual force."""
        return np.array([self.rx, self.ry, self.rz])

    @residual.setter
    def residual(self, value: Sequence[float]):
        [self.rx, self.ry, self.rz] = value

    # supports

    @property
    def support_x(self) -> bool:
        """Get or set the support in x direction."""
        return not self._dof_x.is_active

    @support_x.setter
    def support_x(self, value: bool) -> None:
        self._dof_x.is_active = not value

    @property
    def support_y(self) -> bool:
        """Get or set the support in y direction."""
        return not self._dof_y.is_active

    @support_y.setter
    def support_y(self, value: bool) -> None:
        self._dof_y.is_active = not value

    @property
    def support_z(self) -> bool:
        """Get or set the support in z direction."""
        return not self._dof_z.is_active

    @support_z.setter
    def support_z(self, value: bool) -> None:
        self._dof_z.is_active = not value

    @property
    def support(self) -> str:
        """Get or set the supports."""
        result = ''
        if self.support_x:
            result += 'x'
        if self.support_y:
            result += 'y'
        if self.support_z:
            result += 'z'
        return result

    @support.setter
    def support(self, value: str) -> None:
        self.support_x = 'x' in value
        self.support_y = 'y' in value
        self.support_z = 'z' in value

    # visualization

    def draw(self, item):
        """Draw the node."""

        # reference configuration

        ref_a = self.ref_location.tolist()

        if self.support_x:
            item.append({
                "type": "Support",
                "position": ref_a,
                "direction": "x",
                "layer": "10",
            })

        if self.support_y:
            item.append({
                "type": "Support",
                "position": ref_a,
                "direction": "y",
                "layer": "10",
            })

        if self.support_z:
            item.append({
                "type": "Support",
                "position": ref_a,
                "direction": "z",
                "layer": "10",
            })

        item.append({
            "type": "Points",
            "position": ref_a,
            "material": "PtDarkGray6",
            "layer": "10",
        })

        # actual configuration

        a = self.location.tolist()

        if self.support_x:
            item.append({
                "type": "Support",
                "position": a,
                "direction": "x",
                "layer": "20",
            })

        if self.support_y:
            item.append({
                "type": "Support",
                "position": a,
                "direction": "y",
                "layer": "20",
            })

        if self.support_z:
            item.append({
                "type": "Support",
                "position": a,
                "direction": "z",
                "layer": "20",
            })

        item.append({
            "type": "Points",
            "position": a,
            "material": "PtBlack6",
            "layer": "20",
        })

        item.append({
            "type": "NodeData",
            "position": a,
            "data": {
                "ID": self.id,
                "Location undeformed": self.ref_location.tolist(),
                "Location": self.location.tolist(),
                "Displacement": self.displacement.tolist(),
                "Displacement X": self.u,
                "Displacement Y": self.v,
                "Displacement Z": self.w,
                "Force": self.external_force.tolist(),
                "Force X": self.fx,
                "Force Y": self.fy,
                "Force Z": self.fz,
                "Residual": self.residual.tolist(),
                "Residual X": self.rx,
                "Residual Y": self.ry,
                "Residual Z": self.rz,
            },
        })

        if np.linalg.norm(self.external_force) > 1e-4:
            direction = self.external_force
            direction /= np.linalg.norm(direction)

            item.append({
                "type": "Arrow",
                "position": ref_a,
                "direction": direction.tolist(),
                "color": "DarkGray",
                "layer": "13",
            })

            item.append({
                "type": "Arrow",
                "position": a,
                "direction": direction.tolist(),
                "color": "Green",
                "layer": "23",
            })

        if np.linalg.norm(self.residual) > 1e-4:
            direction = self.residual
            direction /= np.linalg.norm(direction)

            item.append({
                "type": "Arrow",
                "position": a,
                "direction": direction.tolist(),
                "color": "Purple",
                "layer": "23",
            })



"""NFEM teaching tool.

A module for the non linear static analysis of 3D truss problems.
A light weight object oriented FEM code and some usefull postprocessing tools.
"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore

try:
    __version__ = importlib_metadata.version(__name__)
except:
    __version__ = 'dev'

from nfem.model import Model
from nfem.node import Node
from nfem.truss import Truss
from nfem.spring import Spring
from nfem.assembler import Assembler

from nfem.bracketing import bracketing

from nfem.plot import Plot2D

__all__ = [
    'Model',
    'Node',
    'Truss',
    'Spring',
    'Assembler',
    'bracketing',
    'Plot2D',
]



"""Linear spring element."""

from __future__ import annotations

from nfem.dof import Dof
from nfem.node import Node

import numpy as np
import numpy.typing as npt

from typing import Sequence

Vector = npt.NDArray[np.float64]
Matrix = npt.NDArray[np.float64]


class Spring:
    """Linear spring element."""

    def __init__(self, id: str, node: Node, kx: float = 0.0, ky: float = 0.0,
                 kz: float = 0.0):
        """Create a new spring element.

        :id: Unique ID.
        :node: Adjacent node.
        :kx: Stiffness in x direction.
        :ky: Stiffness in y direction.
        :kz: Stiffness in z direction.
        """
        self.id = id
        self.node = node
        self.kx = kx
        self.ky = ky
        self.kz = kz

    @property
    def dofs(self) -> Sequence[Dof]:
        """Get the degrees of freedom."""
        node = self.node
        return [node._dof_x, node._dof_y, node._dof_z]

    # linear analysis

    def compute_linear_r(self) -> Vector:
        """Compute the linear residual force vector of the element."""
        return np.array([
            self.kx * self.node.u,
            self.ky * self.node.v,
            self.kz * self.node.w,
        ], float)

    def compute_linear_k(self) -> Matrix:
        """Compute the linear stiffness matrix of the element."""
        return np.array([
            [self.kx, 0, 0],
            [0, self.ky, 0],
            [0, 0, self.kz],
        ], float)

    def compute_linear_kg(self) -> Matrix:
        """Compute the linear geometric stiffness matrix of the element."""
        return np.zeros((3, 3))

    # nonlinear analysis

    def compute_r(self) -> Vector:
        """Compute the nonlinear residual force vector of the element."""
        return self.compute_linear_r()

    def compute_k(self) -> Matrix:
        """Compute the nonlinear stiffness matrix of the element."""
        return self.compute_ke()

    def compute_ke(self) -> Matrix:
        """Compute the elastic stiffness matrix of the element."""
        return self.compute_linear_k()

    def compute_km(self) -> Matrix:
        """Compute the material stiffness matrix of the element."""
        return self.compute_linear_k()

    def compute_kg(self) -> Matrix:
        """Compute the geometric stiffness matrix of the element."""
        return np.zeros((3, 3))

    def compute_kd(self) -> Matrix:
        """Compute the initial-displacement stiffness matrix of the element."""
        return np.zeros((3, 3))

    # visualization

    def draw(self, canvas) -> None:
        """Draw the spring."""
        location = self.node.location.tolist()

        if self.kx != 0:
            canvas.append({
                'type': 'Spring',
                'position': location,
                'direction': 'x',
            })

        if self.ky != 0:
            canvas.append({
                'type': 'Spring',
                'position': location,
                'direction': 'y',
            })

        if self.kz != 0:
            canvas.append({
                'type': 'Spring',
                'position': location,
                'direction': 'z',
            })



"""Degree of freedom of a mechanical model."""

from __future__ import annotations

from typing import Tuple


class Dof:
    """Degree of freedom of a mechanical model."""

    def __init__(self, id: Tuple[str, str], value: float):
        """Create a new degree of freedom.

        :id: Unique ID.
        :value: Actual value.
        """
        self.id: Tuple[str, str] = id
        self.ref_value: float = value
        self.value: float = value
        self.is_active: bool = True
        self.external_force: float = 0.0
        self.residual: float = 0.0

    def __eq__(self, other) -> bool:
        """Compare for equality.

        :other: Object for comparison.
        :return: True when both objects are equal.
        """
        if isinstance(other, self.__class__):
            return self.id == other.id
        return self.id == other

    def __ne__(self, other) -> bool:
        """Compare for inequality.

        :other: Object for comparison.
        :return: True when both objects are not equal.
        """
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """Hash code for comparisons."""
        return hash(self.id)

    @property
    def delta(self) -> float:
        """Get or set the difference between actual and reference value."""
        return self.value - self.ref_value

    @delta.setter
    def delta(self, value: float) -> None:
        self.value = self.ref_value + value



"""Nonlinear truss element."""

from __future__ import annotations

from nfem.dof import Dof
from nfem.node import Node

import numpy as np
import numpy.linalg as la
import numpy.typing as npt

from typing import Optional, Sequence

Vector = npt.NDArray[np.float64]
Matrix = npt.NDArray[np.float64]

dg = np.array([
    [-1, 0, 0, 1, 0, 0],
    [0, -1, 0, 0, 1, 0],
    [0, 0, -1, 0, 0, 1],
])


def _select(sigma, negative, zero, positive):
    if sigma < -1e-4:
        return negative
    if sigma > 1e-4:
        return positive
    return zero


class Truss:
    """Nonlinear truss element."""

    node_a: Node
    node_b: Node
    youngs_modulus: float
    area: float
    prestress: float
    tensile_strength: Optional[float]
    compressive_strength: Optional[float]

    def __init__(self, id: str, node_a: Node, node_b: Node,
                 youngs_modulus: float, area: float, prestress: float = 0.0):
        """Create a new truss element.

        :node_a: First node.
        :node_b: Second node.
        :youngs_modulus: Young's modulus of the material.
        :area: Area of the cross section.
        :prestress: Internal prestress.
        """
        self.id = id
        self.node_a = node_a
        self.node_b = node_b
        self.youngs_modulus = youngs_modulus
        self.area = area
        self.prestress = prestress
        self.tensile_strength: Optional[float] = None
        self.compressive_strength: Optional[float] = None

    @property
    def dofs(self) -> Sequence[Dof]:
        """Get the degrees of freedom."""
        node_a = self.node_a
        node_b = self.node_b

        return [
            node_a._dof_x, node_a._dof_y, node_a._dof_z,
            node_b._dof_x, node_b._dof_y, node_b._dof_z,
        ]

    @property
    def ref_length(self) -> float:
        """Get the length of the undeformed truss."""
        a = self.node_a.ref_location
        b = self.node_b.ref_location

        return la.norm(b - a)  # type: ignore

    @property
    def length(self) -> float:
        """Get the length of the deformed truss."""
        a = self.node_a.location
        b = self.node_b.location

        return la.norm(b - a)  # type: ignore

    # strain and stress

    def compute_epsilon_gl(self) -> float:
        """Get the Green-Lagrange strain."""
        # reference base vector
        A1 = self.node_b.ref_location - self.node_a.ref_location

        # actual base vector
        a1 = self.node_b.location - self.node_a.location

        return (a1 @ a1) / (2 * A1 @ A1) - 0.5

    def compute_epsilon_lin(self) -> float:
        """Get the linear strain."""
        # reference base vector
        A1 = self.node_b.ref_location - self.node_a.ref_location

        # actual base vector
        a1 = self.node_b.location - self.node_a.location

        return (a1 @ A1) / (A1 @ A1) - 1

    def compute_sigma_pk2(self) -> float:
        """Get the second Piola-Kirchoff stress."""
        eps = self.compute_epsilon_gl()

        return eps * self.youngs_modulus + self.prestress

    @property
    def normal_force(self) -> float:
        """Get normal force."""
        sigma_b = self.compute_sigma_pk2() * self.length / self.ref_length

        return sigma_b * self.area

    # linear analysis

    def compute_linear_r(self) -> Vector:
        """Compute the linear residual force vector of the element."""
        a1 = self.node_b.location - self.node_a.location
        A1 = self.node_b.ref_location - self.node_a.ref_location

        A11 = A1 @ A1
        L = np.sqrt(A11)

        eps = a1 @ A1 / A11 - 1
        sig = eps * self.youngs_modulus + self.prestress

        dp = sig * self.area * L / A11 * A1

        return dp @ dg

    def compute_linear_k(self) -> Matrix:
        """Compute the linear stiffness matrix of the element."""
        A1 = self.node_b.ref_location - self.node_a.ref_location

        A11 = A1 @ A1
        L = np.sqrt(A11)

        ddp = self.youngs_modulus * self.area / A11**2 * L * np.outer(A1, A1)

        return dg.T @ ddp @ dg

    def compute_linear_kg(self) -> Matrix:
        """Compute the linear geometric stiffness matrix of the element."""
        a1 = self.node_b.location - self.node_a.location
        A1 = self.node_b.ref_location - self.node_a.ref_location

        A11 = A1 @ A1
        L = np.sqrt(A11)

        eps = a1 @ A1 / A11 - 1
        sig = eps * self.youngs_modulus + self.prestress

        ddp = sig * self.area / A11 * L * np.eye(3)

        return dg.T @ ddp @ dg

    # nonlinear analysis

    def compute_r(self) -> Vector:
        """Compute the nonlinear residual force vector of the element."""
        a1 = self.node_b.location - self.node_a.location
        A1 = self.node_b.ref_location - self.node_a.ref_location

        A11 = A1 @ A1
        L = np.sqrt(A11)

        eps = 0.5 * (a1 @ a1 / A11 - 1)
        sig = eps * self.youngs_modulus + self.prestress

        dp = sig * self.area * L / A11 * a1

        return dp @ dg

    def compute_k(self) -> Matrix:
        """Compute the nonlinear stiffness matrix of the element."""
        a1 = self.node_b.location - self.node_a.location
        A1 = self.node_b.ref_location - self.node_a.ref_location

        A11 = A1 @ A1
        L = np.sqrt(A11)

        eps = 0.5 * (a1 @ a1 / A11 - 1)
        sig = eps * self.youngs_modulus + self.prestress

        ddp = (
            self.youngs_modulus * self.area / A11**2 * L * np.outer(a1, a1) +
            sig * self.area / A11 * L * np.eye(3)
        )

        return dg.T @ ddp @ dg

    def compute_ke(self) -> Matrix:
        """Compute the elastic stiffness matrix of the element."""
        return self.compute_linear_k()

    def compute_km(self) -> Matrix:
        """Compute the material stiffness matrix of the element."""
        a1 = self.node_b.location - self.node_a.location
        A1 = self.node_b.ref_location - self.node_a.ref_location

        A11 = A1 @ A1
        L = np.sqrt(A11)

        ddp = self.youngs_modulus * self.area / A11**2 * L * np.outer(a1, a1)

        return dg.T @ ddp @ dg

    def compute_kg(self) -> Matrix:
        """Compute the geometric stiffness matrix of the element."""
        a1 = self.node_b.location - self.node_a.location
        A1 = self.node_b.ref_location - self.node_a.ref_location

        A11 = A1 @ A1
        L = np.sqrt(A11)

        eps = 0.5 * (a1 @ a1 / A11 - 1)
        sig = eps * self.youngs_modulus + self.prestress

        ddp = sig * self.area / A11 * L * np.eye(3)

        return dg.T @ ddp @ dg

    def compute_kd(self) -> Matrix:
        """Compute the initial-displacement stiffness matrix of the element."""
        km = self.compute_km()
        ke = self.compute_linear_k()

        return km - ke

    # visualization

    def draw(self, item) -> None:
        """Draw the truss."""

        # reference configuration

        ref_a = self.node_a.ref_location.tolist()
        ref_b = self.node_b.ref_location.tolist()

        item.append({
            "type": "Line",
            "material": "LnDarkGray1",
            "position": [*ref_a, *ref_b],
            "layer": 0,
            "layer": "10",
        })

        # actual configuration

        a = self.node_a.location.tolist()
        b = self.node_b.location.tolist()

        sigma = self.compute_sigma_pk2()

        item.append({
            "type": "Line",
            "material": _select(sigma, "LnBlue2", "LnBlack2", "LnRed2"),
            "position": [*a, *b],
            "layer": 0,
            "layer": "20",
        })

        c = (0.5 * (self.node_a.location + self.node_b.location)).tolist()

        eta = None

        t_strength = self.tensile_strength
        c_strength = self.compressive_strength

        if sigma > 1e-3:
            if t_strength is not None:
                sigma_max = t_strength
                eta = sigma / sigma_max
        elif sigma < -1e-3:
            if c_strength is not None:
                sigma_max = -c_strength
                eta = sigma / sigma_max
        elif t_strength is not None and c_strength is not None:
            eta = 0.0

        item.append({
            "type": "ElementData",
            "position": c,
            "data": {
                "ID": self.id,
                "Length undeformed": self.ref_length,
                "Length": self.length,
                "Engineering Strain": self.compute_epsilon_lin(),
                "Green-Lagrange Strain": self.compute_epsilon_gl(),
                "PK2 Stress": sigma,
                "Normal Force": self.normal_force,
                "Degree of Utilization": eta or "-",
            },
        })



from __future__ import annotations

import numpy as np


def bracketing(model, tol=1e-7, max_steps=100, raise_error=True, **options):
    """Finds next critical point

    The bracketing function finds the next critical point starting from a models
    state.
    - A first nonlinear step needs to be solved already.
    - Det(K) has to be solved for the model.

    1. The algorithm searches for a sign change of Det(K).
        This is done by using a 'last-increment-prediction' and 'arc-length-control'
    2. A bisection algorithm is used to find the Det(K)=0 point.
        Again 'arc-length-control' strategy is used.

    Parameters
    ----------
    model : Model
        initial model for the bracketing function. A first step needs to be already solved
    tol : float
        tolerance for Det(K)==0
    max_steps : int
        maximum number of steps for the iterative search
    options :
        additional options for the nonlinear solution
    """

    print("\n=================================")
    print("Starting bracketing to search for next critical point.")

    if 'solve_attendant_eigenvalue' not in options:
        options['solve_attendant_eigenvalue'] = True
    options['solve_det_k'] = True

    # naming convention indices:
    # _0: current
    # _1: previous
    # _2: second previous

    model_0 = model
    model_1 = model.get_previous_model()
    model_2 = None
    initial_model = model_0

    if model_1 is None:
        raise RuntimeError('One step has to be solved before using the bracketing function.')

    if model_0.det_k is None:
        model_0.solve_det_k()
    det_k_0 = model_0.det_k

    if model_1.det_k is None:
        model_1.solve_det_k()
    det_k_1 = model_1.det_k

    det_k_2 = None

    delta_0 = det_k_0 - det_k_1
    delta_1 = delta_0

    in_min_max = False
    bisectioning = False

    success = False
    step = 0
    while step < max_steps and not success:

        # check if critical point has been found
        if abs(det_k_0) < tol:
            print('\n=================================')
            print('Converged to Det(K) = {}'.format(det_k_0))
            success = True
            break
        elif abs(det_k_0 / initial_model.det_k) < tol:
            print('\n=================================')
            print('Converged to relative value Det(K)/Det(K)_initial = {}'.format(det_k_0 / initial_model.det_k))
            success = True
            break
        elif abs(det_k_0 - det_k_1) < tol:
            print('\n=================================')
            print('WARNING: Converged at stationary point for Det(K)!')
            success = True
            break

        step += 1
        print('\n=================================')
        print('Bracketing step {}'.format(step))

        if not in_min_max and np.sign(det_k_0) == np.sign(det_k_1) and np.sign(delta_0) == np.sign(delta_1):
            print('  Arclength step...')

            model_0 = model_0.get_duplicate()

            model_0.predict_tangential(strategy="arc-length")

            model_0.perform_arc_length_control_step(**options)

        elif bisectioning or np.sign(det_k_0) != np.sign(det_k_1):
            # sign of the determinant changed, target is between det_k_1 and det_k_0
            bisectioning = True
            print('  Bisectioning to find critical point.')
            model_0 = bisection(model_0, **options)

        elif in_min_max or np.sign(delta_0) != np.sign(delta_1):
            # sign of delta changed target is between det_k_2 and det_k_0
            in_min_max = True
            print('  Search for local minimum/maximum.')
            model_0 = minmax(model_0, **options)
        else:
            raise RuntimeError('Unhandled case in bracketing function!')

        # update the local variables
        model_1 = model_0.get_previous_model()
        model_2 = model_1.get_previous_model()

        det_k_2 = model_2.det_k
        det_k_1 = model_1.det_k
        det_k_0 = model_0.det_k

        delta_1 = det_k_1 - det_k_2
        delta_0 = det_k_0 - det_k_1

    if not success:
        msg = 'Bracketing: No critical point found!'
        if raise_error:
            raise RuntimeError(msg)
        else:
            print(msg)
    return model_0


def minmax(model, **options):
    """does a 3-point node search for a local mininmum/maximum. where model is the middle point.
    It returns the estimated position inside model +- arclength"""

    model_1 = model.get_previous_model()
    model_2 = model
    model_3 = model_2.get_duplicate()

    model_3.predict_tangential(strategy="arc-length")

    model_3.perform_arc_length_control_step(**options)

    x1 = 0.0
    x2 = np.linalg.norm(model_2.get_increment_vector())
    x3 = x2 + np.linalg.norm(model_3.get_increment_vector())
    y1 = model_1.det_k
    y2 = model_2.det_k
    y3 = model_3.det_k

    denom = (x1 - x2) * (x1 - x3) * (x2 - x3)
    A = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / denom
    B = (x3*x3 * (y1 - y2) + x2*x2 * (y3 - y1) + x1*x1 * (y2 - y3)) / denom
    # C = (x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3) / denom
    xv = -B / (2 * A)
    # yv = C - B*B / (4*A)

    if xv >= x3:
        return model_3
    elif xv >= x2:
        model = model_2.get_duplicate()

        model.predict_tangential(strategy="arc-length")

        model.scale_prediction((xv-x2)/(x3-x2))

        model.perform_arc_length_control_step(**options)
        return model

    elif xv >= x1:
        model = model_1.get_duplicate()

        model.predict_tangential(strategy="arc-length")

        model.scale_prediction((xv)/(x2))

        model.perform_arc_length_control_step(**options)

        return model
    else:
        print("  WARNING: Minimum/maximum has been already passed!")
        return model


def bisection(model, **options):
    """does a bisectioning to find the root of det_k, between model and its previous model.
    returns the new upper bound"""
    lower_limit_model = model.get_previous_model()
    upper_limit_model = model

    tmp_model = lower_limit_model.get_duplicate()

    for node in tmp_model.nodes:
        lower_node = lower_limit_model.nodes[node.id]
        upper_node = upper_limit_model.nodes[node.id]

        node.u = (lower_node.u + upper_node.u)/2
        node.v = (lower_node.v + upper_node.v)/2
        node.w = (lower_node.w + upper_node.w)/2

    tmp_model.load_factor = (lower_limit_model.load_factor + upper_limit_model.load_factor) / 2

    tmp_model.perform_arc_length_control_step(**options)

    if np.sign(lower_limit_model.det_k) == np.sign(tmp_model.det_k):
        model._previous_model = tmp_model
    else:
        model = tmp_model

    return model



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nfem']

package_data = \
{'': ['*'], 'nfem': ['html/*']}

install_requires = \
['plotly>=5.5.0,<6.0.0']

extras_require = \
{':python_version < "3.10"': ['numpy>=1.20,<2.0', 'scipy>=1.4,<2.0'],
 ':python_version < "3.8"': ['notebook>=5.3,<6.0',
                             'importlib-metadata>=4.11,<5.0',
                             'typing-extensions>=3,<4'],
 ':python_version >= "3.10"': ['numpy>=1.22,<2.0'],
 ':python_version >= "3.10" and python_version < "3.11"': ['scipy>=1.7.2,<2.0.0'],
 ':python_version >= "3.8"': ['notebook>=6.4,<7.0']}

setup_kwargs = {
    'name': 'nfem',
    'version': '4.1.2',
    'description': 'NFEM Teaching Tool',
    'long_description': '# NFEM Teaching Tool\n\n![PyPI](https://img.shields.io/pypi/v/nfem) ![PyPI - Downloads](https://img.shields.io/pypi/dm/nfem) [![Coverage Status](https://coveralls.io/repos/github/StatikTUM/nfem/badge.svg?branch=master&service=github)](https://coveralls.io/github/StatikTUM/nfem?branch=master) ![CI](https://github.com/StatikTUM/nfem/workflows/CI/badge.svg) [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/StatikTUM/nfem/master?filepath=examples) [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/StatikTUM/nfem/)\n\nThis is a lightweight python tool build for teaching non linear FEM analysis of 3D truss structures.\n\n## Installation\n\nInstall [Anaconda](https://www.anaconda.com/distribution/), open the Anaconda console and execute the following command:\n\n```shell\npip install nfem\n```\n',
    'author': 'Thomas Oberbichler',
    'author_email': 'thomas.oberbichler@tum.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/StatikTUM/nfem',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from collections import OrderedDict
from typing import TypeVar, Generic

_KT = TypeVar('_KT')
_VT = TypeVar('_VT')


class KeyCollection(Generic[_KT, _VT]):
    def __init__(self):
        self._dictionary = OrderedDict()

    def __getitem__(self, key: _KT) -> _VT:
        if isinstance(key, int):
            return list(self._dictionary.values())[key]
        return self._dictionary[key]

    def _add(self, value: _VT):
        self._dictionary[value.id] = value

    def __contains__(self, key: _KT) -> bool:
        return self._dictionary.__contains__(key)

    def __len__(self) -> int:
        return self._dictionary.values().__len__()

    def __iter__(self):
        return self._dictionary.values().__iter__()

    def __next__(self):
        return self._dictionary.values().__next__()

    def _ipython_key_completions_(self):
        return list(self._dictionary.keys())
